package medianRanking;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

import misc.MutableInt;
import misc.Pair;
import utils.Debug;
import utils.Wrappers;

/**
 * 
 * @author Bryan BRANCOTTE
 * 
 */
public class MedianRankingTools {

	public static interface Converter<T> {
		public T convert(String elt);
	}

	public static interface Translator<T> {
		public String convert(T elt);
	}

	public static final Converter<String> cvString = new Converter<String>() {
		@Override
		public String convert(String elt) {
			return elt;
		}
	};

	public static final Translator<String> trString = new Translator<String>() {
		@Override
		public String convert(String elt) {
			return elt;
		}
	};

	public static <T> List<Collection<T>> toRankingWithTie(String ranking, Converter<T> converter) {
		LinkedList<Collection<T>> ret = new LinkedList<Collection<T>>();
		Collection<T> bucket;
		int st, en;
		st = ranking.indexOf('[') + 1;
		en = ranking.indexOf(']');
		while (st != -1 && en != -1) {
			bucket = new LinkedList<T>();
			for (String s : ranking.substring(st + 1, en).split(",")) {
				bucket.add(converter.convert(s.trim()));
			}
			ret.add(bucket);
			st = ranking.indexOf('[', en + 1);
			en = ranking.indexOf(']', st + 1);
		}
		return ret;
	}

	public static List<Collection<String>> toRankingWithTie(String ranking) {
		return toRankingWithTie(ranking, new Converter<String>() {
			@Override
			public String convert(String elt) {
				return elt;
			}
		});
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking) {
		return toRankingWithTie(ranking, false);
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking, boolean sortBucket) {
		return toRankingWithTie(ranking, new Translator<T>() {
			public String convert(T elt) {
				return elt.toString();
			};
		}, new ArrayList<T>(), true, sortBucket);
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking, Translator<T> t) {
		return toRankingWithTie(ranking, t, new ArrayList<T>());
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking, Translator<T> t, Collection<T> omega) {
		return toRankingWithTie(ranking, t, omega, true);
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking, Translator<T> t, Collection<T> omega, boolean doSpace) {
		return toRankingWithTie(ranking, t, omega, doSpace, false);
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking, Translator<T> t, Collection<T> omega, boolean doSpace,
			boolean sortBucket) {
		StringBuilder ret = new StringBuilder();
		boolean b1 = false, b2 = false, omegaIteration = false;
		ret.append('[');
		Iterator<Collection<T>> it = ranking.iterator();
		Collection<T> bucket;
		List<String> bucketT = new LinkedList<String>();
		while (it.hasNext() || omega.size() > 0) {
			if (it.hasNext()) {
				bucket = it.next();
			} else {
				omegaIteration = true;
				bucket = omega;
			}
			if (b1) {
				ret.append(',');
				if (doSpace)
					ret.append(' ');
			} else {
				b1 = true;
			}
			ret.append('[');
			if (sortBucket && bucket.size() > 1) {
				for (T elt : bucket) {
					if (!omegaIteration)
						omega.remove(elt);
					bucketT.add(t.convert(elt));
				}
				b2 = false;
				Collections.sort(bucketT);
				for (String elt : bucketT) {
					if (b2) {
						ret.append(',');
						if (doSpace)
							ret.append(' ');
					} else {
						b2 = true;
					}
					ret.append(elt);
				}
			} else {
				b2 = false;
				for (T elt : bucket) {
					if (b2) {
						ret.append(',');
						if (doSpace)
							ret.append(' ');
					} else {
						b2 = true;
					}
					if (!omegaIteration)
						omega.remove(elt);
					ret.append(t.convert(elt));
				}
			}
			ret.append(']');
			if (omegaIteration)
				omega.clear();
		}
		ret.append(']');
		return ret.toString();
	}

	public static <T> String toRankingWithTie(List<Collection<T>> ranking, Translator<T> t, Collection<T> omega, boolean sortBucket,
			String rankingOpener, String rankingCloser, String bucketOpener, String bucketCloser, String elementSeparator,
			String bucketSeparator) {
		StringBuilder ret = new StringBuilder();
		boolean b1 = false, b2 = false, omegaIteration = false;
		ret.append(rankingOpener);
		Iterator<Collection<T>> it = ranking.iterator();
		Collection<T> bucket;
		List<String> bucketT = new LinkedList<String>();
		while (it.hasNext() || omega.size() > 0) {
			if (it.hasNext()) {
				bucket = it.next();
			} else {
				omegaIteration = true;
				bucket = omega;
			}
			if (b1) {
				ret.append(bucketSeparator);
			} else {
				b1 = true;
			}
			ret.append(bucketOpener);
			if (sortBucket && bucket.size() > 1) {
				for (T elt : bucket) {
					if (!omegaIteration)
						omega.remove(elt);
					bucketT.add(t.convert(elt));
				}
				b2 = false;
				Collections.sort(bucketT);
				for (String elt : bucketT) {
					if (b2) {
						ret.append(elementSeparator);
					} else {
						b2 = true;
					}
					ret.append(elt);
				}
			} else {
				b2 = false;
				for (T elt : bucket) {
					if (b2) {
						ret.append(elementSeparator);
					} else {
						b2 = true;
					}
					if (!omegaIteration)
						omega.remove(elt);
					ret.append(t.convert(elt));
				}
			}
			ret.append(bucketCloser);
			if (omegaIteration)
				omega.clear();
		}
		ret.append(rankingCloser);
		return ret.toString();
	}

	public static Collection<Pair<String, List<Collection<String>>>> toRankingWithTieFromFile(String path) {
		return toRankingWithTieFromFile(new File(path));
	}

	public static Collection<Pair<String, List<Collection<String>>>> toRankingWithTieFromFile(File file) {
		if (file.getName().substring(file.getName().lastIndexOf('.') + 1).equals("election"))
			return toRankingWithTieFromElectionFile(file, new Converter<String>() {
				@Override
				public String convert(String elt) {
					return elt;
				}
			}, null);
		return toRankingWithTieFromFile(file, true, new Converter<String>() {
			@Override
			public String convert(String elt) {
				return elt;
			}
		});
	}

	/**
	 * Load rankings in the specified file, if
	 * <b>ignoreGoldStandardIfPresent</b> is true, then we ignore the gold
	 * standard results a.k.a GS. Files have this shape : <br/>
	 * <div> PR:=[[1, 2], [3], [4]]<br/>
	 * IE:=[[2], [4], [3], [1]]<br/>
	 * PC:=[[2], [4], [3], [1]]<br/>
	 * GS:=[[1], [4], [2], [3]]</div>
	 * 
	 * @param path
	 * @param ignoreGoldStandardIfPresent
	 * @param c
	 *            the converter to transform the element to a specified type T
	 * @return
	 */
	public static <T> List<Pair<String, List<Collection<T>>>> toRankingWithTieFromFile(String path, boolean ignoreGoldStandardIfPresent,
			Converter<T> c) {
		return toRankingWithTieFromFile(new File(path), ignoreGoldStandardIfPresent, c);
	}

	/**
	 * Load rankings in the specified file, if
	 * <b>ignoreGoldStandardIfPresent</b> is true, then we ignore the gold
	 * standard results a.k.a GS. Files have this shape : <br/>
	 * <div> PR:=[[1, 2], [3], [4]]<br/>
	 * IE:=[[2], [4], [3], [1]]<br/>
	 * PC:=[[2], [4], [3], [1]]<br/>
	 * GS:=[[1], [4], [2], [3]]</div>
	 * 
	 * @param path
	 * @param ignoreGoldStandardIfPresent
	 * @param the
	 *            converter to transform the element to a specified type T
	 * @return
	 */
	public static <T> List<Pair<String, List<Collection<T>>>> toRankingWithTieFromFile(File file, boolean ignoreGoldStandardIfPresent,
			Converter<T> c) {
		List<Pair<String, List<Collection<T>>>> ret = new LinkedList<Pair<String, List<Collection<T>>>>();
		toRankingWithTieFromFile(file, ignoreGoldStandardIfPresent, c, ret);
		return ret;
	}

	/**
	 * Load rankings in the specified file in the list ret in parameter, if
	 * <b>ignoreGoldStandardIfPresent</b> is true, then we ignore the gold
	 * standard results a.k.a GS. Files have this shape : <br/>
	 * <div> PR:=[[1, 2], [3], [4]]<br/>
	 * IE:=[[2], [4], [3], [1]]<br/>
	 * PC:=[[2], [4], [3], [1]]<br/>
	 * GS:=[[1], [4], [2], [3]]</div>
	 * 
	 * @param path
	 * @param ignoreGoldStandardIfPresent
	 * @param c
	 *            the converter to transform the element to a specified type T
	 * @param ret
	 *            the list which will contain the rankings, if null a new one is
	 *            created
	 * @return
	 */
	public static <T> List<Pair<String, List<Collection<T>>>> toRankingWithTieFromFile(File file, boolean ignoreGoldStandardIfPresent,
			Converter<T> c, List<Pair<String, List<Collection<T>>>> ret) {
		assert file.exists() : "File is missing >" + file + "<";
		if (ret == null)
			ret = new LinkedList<Pair<String, List<Collection<T>>>>();
		BufferedReader br = null;
		try {
			if (file.getAbsolutePath().endsWith(".gz") || file.getAbsolutePath().endsWith(".zip")) {
				// is = new GZIPInputStream(new FileInputStream(file));
				br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file))));
			} else {
				br = new BufferedReader(new FileReader(file));
			}
			String l, lprev = null;
			while ((l = br.readLine()) != null) {
				if (lprev != null) {
					l = lprev + l;
					lprev = null;
				}
				if (l.endsWith("\\")) {// MuPAD's output robustness.
					lprev = l.substring(0, l.length() - 1);
				} else if (!l.startsWith("%")
						&& (!ignoreGoldStandardIfPresent || !l.startsWith("GS_c:=") && !l.startsWith("GoldStandard:=")
								&& !l.startsWith("GS:=")) && l.length() > 0) {
					int pos = l.indexOf('=');
					ret.add(new Pair<String, List<Collection<T>>>(pos == -1 ? "" : l.substring(0, pos - 1), toRankingWithTie(pos == -1 ? l
							: l.substring(pos + 1), c)));
				}
			}
		} catch (FileNotFoundException e) {
			Debug.errln(e);
		} catch (IOException e) {
			Debug.errln(e);
		} finally {
			try {
				if (br != null)
					br.close();
			} catch (IOException e) {
				Debug.errln(e);
			}
		}
		return ret;
	}

	public static List<List<Collection<Integer>>> toRankingWithTieFromElectionFile(File file) {
		return MedianRankingTools.retainOnlyRanking(MedianRankingTools.toRankingWithTieFromElectionFile(file, new Converter<Integer>() {
			Map<String, Integer> map = new HashMap<String, Integer>();

			@Override
			public Integer convert(String elt) {
				Integer i = map.get(elt);
				if (i == null)
					map.put(elt, i = map.size() + 1);
				return i;
			}
		}, null));
	}

	public static <T> List<Pair<String, List<Collection<T>>>> toRankingWithTieFromElectionFile(File file, Converter<T> c,
			List<Pair<String, List<Collection<T>>>> ret) {
		assert file.exists() : "File is missing >" + file + "<";
		if (ret == null)
			ret = new LinkedList<Pair<String, List<Collection<T>>>>();
		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(file));
			String l, lprev = null;
			while ((l = br.readLine()) != null) {
				if (lprev != null) {
					l = lprev + l;
					lprev = null;
				}

				String[] split = l.split(">");
				List<Collection<T>> ranking = new Vector<Collection<T>>(split.length);
				ret.add(new Pair<String, List<Collection<T>>>("", ranking));
				for (String e : split) {
					ranking.add(Wrappers.wrapInCollection(c.convert(e)));
				}
			}
		} catch (FileNotFoundException e) {
			Debug.errln(e);
		} catch (IOException e) {
			Debug.errln(e);
		} finally {
			try {
				if (br != null)
					br.close();
			} catch (IOException e) {
				Debug.errln(e);
			}
		}
		return ret;
	}

	/**
	 * Read a file created from The ImplementationTest process, and create the
	 * ranking with tie corresponding to the results.Files are tsv like : <br/>
	 * <div> ELEMENT_66 \t 0.35<br/>
	 * ELEMENT_54 \t 0.32<br/>
	 * ELEMENT_42 \t 0.32<br/>
	 * ELEMENT_3 \t 0.28<br/>
	 * ...</div><br/>
	 * Elements in those files are ordered in a decreasing order of the grade,
	 * when two following elements have the same grade it implies that they are
	 * in the same bucket.
	 * 
	 * @param path
	 * @param maxElement
	 *            if > 0 then we only keep the n first results
	 * @param endBucket
	 *            if true then the ranking returned will contain element as
	 *            rank(elt)<=50 OR grade(elt)==grade(elt50) AND rank(elt50)=50
	 * @return
	 */
	public static List<Collection<String>> toRankingWithTieFromFRLogFile(String path, int maxElement, boolean endBucket,
			int absolutMaxElement) {
		LinkedList<Collection<String>> ret = new LinkedList<Collection<String>>();
		Collection<String> bucket = null;
		int i = 0;
		float lastGrade = Float.NaN, grade;
		boolean lvlMode = false;
		absolutMaxElement = (maxElement > absolutMaxElement ? maxElement : absolutMaxElement);
		BufferedReader br = null;
		try {
			br = new BufferedReader(new FileReader(new File(path)));
			String l;
			while ((l = br.readLine()) != null) {
				int pos = l.indexOf('\t');
				grade = Float.parseFloat(l.substring(pos + 1));
				if (absolutMaxElement > 0 && i >= absolutMaxElement)
					break;
				if (grade != lastGrade) {
					if (maxElement > 0 && i >= maxElement)
						break;
					if (bucket == null) {
						lvlMode = l.indexOf("_Lvl") != -1;
					}
					bucket = new LinkedList<String>();
					ret.add(bucket);
					lastGrade = grade;
				}
				bucket.add(l.substring(0, lvlMode ? l.indexOf("_Lvl") : pos));
				i++;
			}
		} catch (FileNotFoundException e) {
			// Debug.errln(e);
		} catch (IOException e) {
			Debug.errln(e);
		} finally {
			if (br != null)
				try {
					br.close();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
		}
		return ret;
	}

	public static void toStringRankingWithTie(int[][] rankingWithTie) {
		for (int ib = 0; ib < rankingWithTie.length; ib++) {
			if (ib != 0)
				System.out.print(',');
			System.out.print('[');
			for (int ie = 0; ie < rankingWithTie[ib].length; ie++) {
				if (ie != 0)
					System.out.print(',');
				System.out.print(rankingWithTie[ib][ie]);
			}
			System.out.print(']');
		}
		System.out.print('\n');
	}

	/**
	 * Select all the element at a rank &le; k or in the same bucket as an ohter
	 * elements which is at a rank &le; k<br/>
	 * Let E be all the elements in <b>rankings</b> T ={e \in E, rank(e)\le topK
	 * OR \exists e' \in E AND rank(e')\le topK AND bucket(e)=bucket(e')}
	 * 
	 * @param <T>
	 * @param rankings
	 * @param topK
	 * @return T
	 */
	public static <T> Collection<T> getTopKElements(Collection<List<Collection<T>>> rankings, int topK) {
		HashSet<T> elements = new HashSet<T>();
		for (List<Collection<T>> ranking : rankings) {
			getTopKElements(ranking, topK, elements);
		}
		return elements;
	}

	public static <T> Collection<T> getTopKElements(List<Collection<T>> ranking, int topK, HashSet<T> elements) {
		int t = topK;
		Iterator<Collection<T>> it = ranking.iterator();
		while (it.hasNext() && t > 0) {
			Collection<T> bucket = it.next();
			t -= bucket.size();
			elements.addAll(bucket);
		}
		return elements;
	}

	public static <T> Collection<T> getTopKElements(List<Collection<T>> ranking, int topK) {
		return getTopKElements(ranking, topK, new HashSet<T>());
	}

	public static <T> List<Collection<T>> getTopKRanking(final List<Collection<T>> ranking, final int topK) {
		List<Collection<T>> subRanking = new LinkedList<Collection<T>>();
		int t = topK;
		Iterator<Collection<T>> it = ranking.iterator();
		while (it.hasNext() && t > 0) {
			Collection<T> bucket = it.next();
			subRanking.add(new Vector<T>(bucket));
			t -= bucket.size();
		}
		return subRanking;
	}

	public static <T> Collection<List<Collection<T>>> getTopKRankingsUnified(Collection<List<Collection<T>>> rankings, final int topK) {
		HashSet<T> omega = new HashSet<T>((topK < 0 ? rankings.iterator().next().size() : topK) * rankings.size() << 1);
		Collection<List<Collection<T>>> subRankings = new Vector<List<Collection<T>>>(rankings.size());
		for (List<Collection<T>> ranking : rankings) {
			List<Collection<T>> subRanking = new LinkedList<Collection<T>>();
			int t = topK;
			Iterator<Collection<T>> it = ranking.iterator();
			while (it.hasNext() && (topK < 0 || t > 0)) {
				Collection<T> bucket = it.next();
				omega.addAll(bucket);
				subRanking.add(new Vector<T>(bucket));
				t -= bucket.size();
			}
			subRankings.add(subRanking);
		}
		return projectRankings(subRankings, omega, true);
	}

	/**
	 * Project the ranking over only the elements of <b>elementsToConsider<b/> :
	 * projectRankings("[[1,2],[3],[4,5]]","{1,2,4,6,7}")=
	 * <ul>
	 * <li>addMissingOneAtTheEnd==true <=>"[[1,2],[4],[6,7]]"</li>
	 * <li>addMissingOneAtTheEnd==false<=>"[[1,2],[4]]"</li>
	 * </ul>
	 * 
	 * @param <T>
	 * @param rankings
	 * @param elementsToConsider
	 * @param addMissingOneAtTheEnd
	 * @return
	 */
	public static <T> List<List<Collection<T>>> projectRankings(Collection<List<Collection<T>>> rankings, Collection<T> elementsToConsider,
			boolean addMissingOneAtTheEnd) {
		List<List<Collection<T>>> rankingsOut = new LinkedList<List<Collection<T>>>();
		for (List<Collection<T>> ranking : rankings) {
			List<Collection<T>> rankingOut = new LinkedList<Collection<T>>();
			Collection<T> omega = new HashSet<T>(elementsToConsider);// omega=new
																		// java.util.HashSet<T>(omega)
			Iterator<Collection<T>> it = ranking.iterator();
			while (it.hasNext() && omega.size() > 0) {
				Collection<T> bucket = (Collection<T>) it.next();
				Collection<T> bucketOut = new Vector<T>(bucket);
				bucketOut.retainAll(omega);
				omega.removeAll(bucket);
				if (bucketOut.size() > 0)
					rankingOut.add(bucketOut);
			}
			if (addMissingOneAtTheEnd && omega.size() > 0)
				rankingOut.add(omega);
			if (rankingOut.size() > 0)
				rankingsOut.add(rankingOut);
		}
		return rankingsOut;
	}

	public static <T> List<List<Collection<T>>> retainOnlyRanking(Collection<Pair<String, List<Collection<T>>>> rankingAndNames) {
		List<List<Collection<T>>> rankings = new LinkedList<List<Collection<T>>>();
		for (Pair<String, List<Collection<T>>> p : rankingAndNames) {
			rankings.add(p.getParam2());
		}
		return rankings;
	}

	public static <T> Set<T> getDistinctElement(Collection<List<Collection<T>>> rankings, Set<T> set) {
		if (set == null)
			set = new HashSet<T>();
		for (List<Collection<T>> ranking : rankings)
			for (Collection<T> bucket : ranking)
				set.addAll(bucket);
		return set;
	}

	public static <T> Set<T> getDistinctElement(Collection<List<Collection<T>>> rankings) {
		return getDistinctElement(rankings, new HashSet<T>());
	}

	public static <T> Set<T> getDistinctElement(Collection<List<Collection<T>>> rankings, int inAtLeastNRankings) {
		HashMap<T, MutableInt> map = new HashMap<T, MutableInt>();
		for (List<Collection<T>> ranking : rankings)
			for (Collection<T> bucket : ranking)
				for (T t : bucket) {
					MutableInt mi = map.get(t);
					if (mi == null)
						map.put(t, new MutableInt(1));
					else
						mi.i++;
				}
		Set<T> res = new HashSet<T>();
		Iterator<Entry<T, MutableInt>> it = map.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<T, misc.MutableInt> entry = (Map.Entry<T, misc.MutableInt>) it.next();
			if (entry.getValue().i >= inAtLeastNRankings)
				res.add(entry.getKey());
		}
		return res;
	}

	public static <T> int getRankingLength(List<Collection<T>> r) {
		int ret = 0;
		for (Collection<T> b : r)
			ret += b.size();
		return ret;
	}

	public static <T> List<Collection<T>> toRankingWithTiesFromPermutation(List<T> permutation) {
		List<Collection<T>> ranking = new Vector<Collection<T>>(permutation.size());
		for (T e : permutation) {
			ranking.add(Wrappers.wrapInCollection(e));
		}
		return ranking;
	}
}
